#!/usr/bin/python

# (c) 2019, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
author: NetApp Ansible Team (@carchi8py) <ng-ansibleteam@netapp.com>
description:
  - Update ONTAP service-prosessor firmware
  - The recommend procedure is to
    1. download the firmware package from the NetApp Support site
    2. copy the package to a web server
    3. download the package from the web server using this module
  - Once a disk qualification, disk, shelf, or ACP firmware package is downloaded, ONTAP will automatically update the related resources in background.
  - It may take some time to complete.
  - For service processor, the update requires a node reboot to take effect.
extends_documentation_fragment:
  - netapp.ontap.netapp.na_ontap
module: na_ontap_firmware_upgrade
options:
  state:
    description:
      - Whether the specified ONTAP firmware should  be upgraded or not.
    default: present
    type: str
  node:
    description:
      - Node on which the device is located.
      - Not required if package_url is present and force_disruptive_update is False.
      - If this option is not given, the firmware will be downloaded on all nodes in the cluster,
      - and the resources will be updated in background on all nodes, except for service processor.
      - For service processor, the upgrade will happen automatically when each node is rebooted.
    type: str
  clear_logs:
    description:
      - Clear logs on the device after update. Default value is true.
      - Not used if force_disruptive_update is False.
    type: bool
    default: true
  package:
    description:
      - Name of the package file containing the firmware to be installed. Not required when -baseline is true.
      - Not used if force_disruptive_update is False.
    type: str
  package_url:
    description:
      - URL of the package file containing the firmware to be downloaded.
      - Once the package file is downloaded to a node, the firmware update will happen automatically in background.
      - For SP, the upgrade will happen automatically when a node is rebooted.
      - For SP, the upgrade will happen automatically if autoupdate is enabled (which is the recommended setting).
    version_added: "20.4.1"
    type: str
  force_disruptive_update:
    description:
      - If set to C(False), and URL is given, the upgrade is non disruptive. If URL is not given, no operation is performed.
      - Do not set this to C(True), unless directed by NetApp Tech Support.
      - It will force an update even if the resource is not ready for it, and can be disruptive.
    type: bool
    version_added: "20.4.1"
    default: False
  shelf_module_fw:
    description:
      - Shelf module firmware to be updated to.
      - Not used if force_disruptive_update is False (ONTAP will automatically select the firmware)
    type: str
  disk_fw:
    description:
      - disk firmware to be updated to.
      - Not used if force_disruptive_update is False (ONTAP will automatically select the firmware)
    type: str
  update_type:
    description:
      - Type of firmware update to be performed. Options include serial_full, serial_differential, network_full.
      - Not used if force_disruptive_update is False (ONTAP will automatically select the firmware)
    type: str
  install_baseline_image:
    description:
      - Install the version packaged with ONTAP if this parameter is set to true. Otherwise, package must be used to specify the package to install.
      - Not used if force_disruptive_update is False (ONTAP will automatically select the firmware)
    type: bool
    default: false
  firmware_type:
    description:
      - Type of firmware to be upgraded. Options include shelf, ACP, service-processor, and disk.
      - For shelf firmware upgrade the operation is asynchronous, and therefore returns no errors that might occur during the download process.
      - Shelf firmware upgrade is idempotent if shelf_module_fw is provided .
      - disk firmware upgrade is idempotent if disk_fw is provided .
      - With check mode, SP, ACP, disk, and shelf firmware upgrade is not idempotent.
      - This operation will only update firmware on shelves/disk that do not have the latest firmware-revision.
      - Not used if force_disruptive_update is False (ONTAP will automatically detect the firmware type)
    choices: ['service-processor', 'shelf', 'acp', 'disk']
    type: str
  fail_on_502_error:
    description:
      - The firmware download may take time if the web server is slow and if there are many nodes in the cluster.
      - ONTAP will break the ZAPI connection after 5 minutes with a 502 Bad Gateway error, even though the download \
is still happening.
      - By default, this module ignores this error and assumes the download is progressing as ONTAP does not \
provide a way to check the status.
      - When setting this option to true, the module will report 502 as an error.
    type: bool
    default: false
    version_added: "20.6.0"
  rename_package:
    description:
      - Rename the package.
      - Only available if 'firmware_type' is 'service-processor'.
    type: str
    version_added: "20.6.1"
  replace_package:
    description:
      - Replace the local package.
      - Only available if 'firmware_type' is 'service-processor'.
    type: bool
    version_added: "20.6.1"
  reboot_sp:
    description:
      - Reboot service processor before downloading package.
      - Only available if 'firmware_type' is 'service-processor'.
    type: bool
    default: true
    version_added: "20.6.1"
short_description:  NetApp ONTAP firmware upgrade for SP, shelf, ACP, and disk.
version_added: 2.9.0
'''

EXAMPLES = """

    - name: firmware upgrade
      na_ontap_firmware_upgrade:
        state: present
        package_url: "{{ web_link }}"
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
    - name: firmware upgrade, confirm successful download
      na_ontap_firmware_upgrade:
        state: present
        package_url: "{{ web_link }}"
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
        fail_on_502_error: true
    - name: SP firmware upgrade
      na_ontap_firmware_upgrade:
        state: present
        node: vsim1
        package: "{{ file name }}"
        package_url: "{{ web_link }}"
        clear_logs: True
        install_baseline_image: False
        update_type: serial_full
        force_disruptive_update: False
        firmware_type: service-processor
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
    - name: SP firmware download replace package
      tags:
      - sp_download
      na_ontap_firmware_upgrade:
        state: present
        node: vsim1
        package_url: "{{ web_link }}"
        firmware_type: service-processor
        replace_package: true
        reboot_sp: true
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
        https: true
        validate_certs: false
    - name: SP firmware download rename package
      tags:
      - sp_download
      na_ontap_firmware_upgrade:
        state: present
        node: vsim1
        package_url: "{{ web_link }}"
        firmware_type: service-processor
        rename_package: SP_FW.zip
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
        https: true
        validate_certs: false
    - name: ACP firmware download and upgrade
      na_ontap_firmware_upgrade:
        state: present
        node: vsim1
        firmware_type: acp
        force_disruptive_update: False
        package_url: "{{ web_link }}"
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
    - name: shelf firmware upgrade
      na_ontap_firmware_upgrade:
        state: present
        firmware_type: shelf
        shelf_module_fw: 1221
        force_disruptive_update: False
        package_url: "{{ web_link }}"
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
    - name: disk firmware upgrade
      na_ontap_firmware_upgrade:
        state: present
        firmware_type: disk
        disk_fw: NA02
        force_disruptive_update: False
        package_url: "{{ web_link }}"
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
"""

RETURN = """
msg:
    description: Returns additional information in case of success.
    returned: always
    type: str
"""

import traceback
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
import ansible_collections.netapp.ontap.plugins.module_utils.netapp as netapp_utils
from ansible_collections.netapp.ontap.plugins.module_utils.netapp_module import NetAppModule
import time


HAS_NETAPP_LIB = netapp_utils.has_netapp_lib()
MSGS = dict(
    no_action='No action taken.',
    dl_completed='Firmware download completed.',
    dl_completed_slowly='Firmware download completed, slowly.',
    dl_in_progress='Firmware download still in progress.'
)


class NetAppONTAPFirmwareUpgrade(object):
    """
    Class with ONTAP firmware upgrade methods
    """

    def __init__(self):
        self.argument_spec = netapp_utils.na_ontap_host_argument_spec()
        self.argument_spec.update(dict(
            state=dict(required=False, type='str', default='present'),
            node=dict(required=False, type='str'),
            firmware_type=dict(required=False, type='str', choices=['service-processor', 'shelf', 'acp', 'disk']),
            clear_logs=dict(required=False, type='bool', default=True),
            package=dict(required=False, type='str'),
            install_baseline_image=dict(required=False, type='bool', default=False),
            update_type=dict(required=False, type='str'),
            shelf_module_fw=dict(required=False, type='str'),
            disk_fw=dict(required=False, type='str'),
            package_url=dict(required=False, type='str'),
            force_disruptive_update=dict(required=False, type='bool', default=False),
            fail_on_502_error=dict(required=False, type='bool', default=False),
            rename_package=dict(required=False, type='str'),
            replace_package=dict(required=False, type='bool'),
            reboot_sp=dict(required=False, type='bool', default=True)
        ))

        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            required_if=[
                ('firmware_type', 'acp', ['node']),
                ('firmware_type', 'disk', ['node']),
                ('firmware_type', 'service-processor', ['node']),
                ('force_disruptive_update', True, ['firmware_type']),
            ],
            supports_check_mode=True
        )

        self.na_helper = NetAppModule()
        self.parameters = self.na_helper.set_parameters(self.module.params)
        if self.parameters.get('firmware_type') == 'service-processor':
            if self.parameters.get('install_baseline_image') and self.parameters.get('package') is not None:
                self.module.fail_json(msg='Do not specify both package and install_baseline_image: true')
            if not self.parameters.get('package') and self.parameters.get('install_baseline_image') == 'False':
                self.module.fail_json(msg='Specify at least one of package or install_baseline_image')
        if HAS_NETAPP_LIB is False:
            self.module.fail_json(msg="the python NetApp-Lib module is required")
        else:
            self.server = netapp_utils.setup_na_ontap_zapi(module=self.module, wrap_zapi=True)

    def firmware_image_get_iter(self):
        """
        Compose NaElement object to query current firmware version
        :return: NaElement object for firmware_image_get_iter with query
        """
        firmware_image_get = netapp_utils.zapi.NaElement('service-processor-get-iter')
        query = netapp_utils.zapi.NaElement('query')
        firmware_image_info = netapp_utils.zapi.NaElement('service-processor-info')
        firmware_image_info.add_new_child('node', self.parameters['node'])
        query.add_child_elem(firmware_image_info)
        firmware_image_get.add_child_elem(query)
        return firmware_image_get

    def firmware_image_get(self, node_name):
        """
        Get current firmware image info
        :return: True if query successful, else return None
        """
        firmware_image_get_iter = self.firmware_image_get_iter()
        try:
            result = self.server.invoke_successfully(firmware_image_get_iter, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching firmware image details: %s: %s'
                                      % (self.parameters['node'], to_native(error)),
                                  exception=traceback.format_exc())
        # return firmware image details
        if result.get_child_by_name('num-records') and int(result.get_child_content('num-records')) > 0:
            sp_info = result.get_child_by_name('attributes-list').get_child_by_name('service-processor-info')
            firmware_version = sp_info.get_child_content('firmware-version')
            return firmware_version
        return None

    def acp_firmware_required_get(self):
        """
        where acp firmware upgrade is required
        :return:  True is firmware upgrade is required else return None
        """
        acp_firmware_get_iter = netapp_utils.zapi.NaElement('storage-shelf-acp-module-get-iter')
        query = netapp_utils.zapi.NaElement('query')
        acp_info = netapp_utils.zapi.NaElement('storage-shelf-acp-module')
        query.add_child_elem(acp_info)
        acp_firmware_get_iter.add_child_elem(query)
        try:
            result = self.server.invoke_successfully(acp_firmware_get_iter, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching acp firmware details details: %s'
                                  % (to_native(error)), exception=traceback.format_exc())
        if result.get_child_by_name('attributes-list').get_child_by_name('storage-shelf-acp-module'):
            acp_module_info = result.get_child_by_name('attributes-list').get_child_by_name(
                'storage-shelf-acp-module')
            state = acp_module_info.get_child_content('state')
            if state == 'firmware_update_required':
                # acp firmware version upgrade required
                return True
        return False

    def sp_firmware_image_update_progress_get(self, node_name):
        """
        Get current firmware image update progress info
        :return: Dictionary of firmware image update progress if query successful, else return None
        """
        firmware_update_progress_get = netapp_utils.zapi.NaElement('service-processor-image-update-progress-get')
        firmware_update_progress_get.add_new_child('node', self.parameters['node'])

        firmware_update_progress_info = dict()
        try:
            result = self.server.invoke_successfully(firmware_update_progress_get, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching firmware image upgrade progress details: %s'
                                      % (to_native(error)), exception=traceback.format_exc())
        # return firmware image update progress details
        if result.get_child_by_name('attributes').get_child_by_name('service-processor-image-update-progress-info'):
            update_progress_info = result.get_child_by_name('attributes').get_child_by_name('service-processor-image-update-progress-info')
            firmware_update_progress_info['is-in-progress'] = update_progress_info.get_child_content('is-in-progress')
            firmware_update_progress_info['node'] = update_progress_info.get_child_content('node')
        return firmware_update_progress_info

    def shelf_firmware_info_get(self):
        """
        Get the current firmware of shelf module
        :return:dict with module id and firmware info
        """
        shelf_id_fw_info = dict()
        shelf_firmware_info_get = netapp_utils.zapi.NaElement('storage-shelf-info-get-iter')
        desired_attributes = netapp_utils.zapi.NaElement('desired-attributes')
        storage_shelf_info = netapp_utils.zapi.NaElement('storage-shelf-info')
        shelf_module = netapp_utils.zapi.NaElement('shelf-modules')
        shelf_module_info = netapp_utils.zapi.NaElement('storage-shelf-module-info')
        shelf_module.add_child_elem(shelf_module_info)
        storage_shelf_info.add_child_elem(shelf_module)
        desired_attributes.add_child_elem(storage_shelf_info)
        shelf_firmware_info_get.add_child_elem(desired_attributes)

        try:
            result = self.server.invoke_successfully(shelf_firmware_info_get, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching shelf module firmware  details: %s'
                                      % (to_native(error)), exception=traceback.format_exc())
        if result.get_child_by_name('num-records') and int(result.get_child_content('num-records')) > 0:
            shelf_info = result.get_child_by_name('attributes-list').get_child_by_name('storage-shelf-info')
            if (shelf_info.get_child_by_name('shelf-modules') and
                    shelf_info.get_child_by_name('shelf-modules').get_child_by_name('storage-shelf-module-info')):
                shelves = shelf_info['shelf-modules'].get_children()
                for shelf in shelves:
                    shelf_id_fw_info[shelf.get_child_content('module-id')] = shelf.get_child_content('module-fw-revision')
        return shelf_id_fw_info

    def disk_firmware_info_get(self):
        """
        Get the current firmware of disks module
        :return:
        """
        disk_id_fw_info = dict()
        disk_firmware_info_get = netapp_utils.zapi.NaElement('storage-disk-get-iter')
        desired_attributes = netapp_utils.zapi.NaElement('desired-attributes')
        storage_disk_info = netapp_utils.zapi.NaElement('storage-disk-info')
        disk_inv = netapp_utils.zapi.NaElement('disk-inventory-info')
        storage_disk_info.add_child_elem(disk_inv)
        desired_attributes.add_child_elem(storage_disk_info)
        disk_firmware_info_get.add_child_elem(desired_attributes)
        try:
            result = self.server.invoke_successfully(disk_firmware_info_get, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching disk module firmware  details: %s'
                                      % (to_native(error)), exception=traceback.format_exc())
        if result.get_child_by_name('num-records') and int(result.get_child_content('num-records')) > 0:
            disk_info = result.get_child_by_name('attributes-list')
            disks = disk_info.get_children()
            for disk in disks:
                disk_id_fw_info[disk.get_child_content('disk-uid')] = disk.get_child_by_name('disk-inventory-info').get_child_content('firmware-revision')
        return disk_id_fw_info

    def disk_firmware_required_get(self):
        """
        Check weather disk firmware upgrade is required or not
        :return: True if the firmware upgrade is required
        """
        disk_firmware_info = self.disk_firmware_info_get()
        for disk in disk_firmware_info:
            if (disk_firmware_info[disk]) != self.parameters['disk_fw']:
                return True
        return False

    def shelf_firmware_required_get(self):
        """
        Check weather shelf firmware upgrade is required or not
        :return: True if the firmware upgrade is required
        """
        shelf_firmware_info = self.shelf_firmware_info_get()
        for module in shelf_firmware_info:
            if (shelf_firmware_info[module]) != self.parameters['shelf_module_fw']:
                return True
        return False

    def sp_firmware_image_update(self):
        """
        Update current firmware image
        """
        firmware_update_info = netapp_utils.zapi.NaElement('service-processor-image-update')
        if self.parameters.get('package') is not None:
            firmware_update_info.add_new_child('package', self.parameters['package'])
        if self.parameters.get('clear_logs') is not None:
            firmware_update_info.add_new_child('clear-logs', str(self.parameters['clear_logs']))
        if self.parameters.get('install_baseline_image') is not None:
            firmware_update_info.add_new_child('install-baseline-image', str(self.parameters['install_baseline_image']))
        firmware_update_info.add_new_child('node', self.parameters['node'])
        firmware_update_info.add_new_child('update-type', self.parameters['update_type'])

        try:
            self.server.invoke_successfully(firmware_update_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            # Current firmware version matches the version to be installed
            if to_native(error.code) == '13001' and (error.message.startswith('Service Processor update skipped')):
                return False
            self.module.fail_json(msg='Error updating firmware image for %s: %s'
                                      % (self.parameters['node'], to_native(error)),
                                  exception=traceback.format_exc())
        return True

    def shelf_firmware_upgrade(self):
        """
        Upgrade shelf firmware image
        """
        shelf_firmware_update_info = netapp_utils.zapi.NaElement('storage-shelf-firmware-update')
        try:
            self.server.invoke_successfully(shelf_firmware_update_info, enable_tunneling=True)
            return True
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error updating shelf firmware image : %s'
                                      % (to_native(error)), exception=traceback.format_exc())

    def acp_firmware_upgrade(self):

        """
        Upgrade shelf firmware image
        """
        acp_firmware_update_info = netapp_utils.zapi.NaElement('storage-shelf-acp-firmware-update')
        acp_firmware_update_info.add_new_child('node-name', self.parameters['node'])
        try:
            self.server.invoke_successfully(acp_firmware_update_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error updating acp firmware image : %s'
                                  % (to_native(error)), exception=traceback.format_exc())

    def disk_firmware_upgrade(self):

        """
        Upgrade disk firmware
        """
        disk_firmware_update_info = netapp_utils.zapi.NaElement('disk-update-disk-fw')
        disk_firmware_update_info.add_new_child('node-name', self.parameters['node'])
        try:
            self.server.invoke_successfully(disk_firmware_update_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error updating disk firmware image : %s'
                                  % (to_native(error)), exception=traceback.format_exc())
        return True

    def download_firmware(self):
        ''' calls the system-cli ZAPI as there is no ZAPI for this feature '''
        msg = MSGS['dl_completed']
        command = ['storage', 'firmware', 'download', '-node', self.parameters['node'] if self.parameters.get('node') else '*',
                   '-package-url', self.parameters['package_url']]
        command_obj = netapp_utils.zapi.NaElement("system-cli")

        args_obj = netapp_utils.zapi.NaElement("args")
        for arg in command:
            args_obj.add_new_child('arg', arg)
        command_obj.add_child_elem(args_obj)
        command_obj.add_new_child('priv', 'advanced')

        output = None
        try:
            output = self.server.invoke_successfully(command_obj, True)

        except netapp_utils.zapi.NaApiError as error:
            # with nettap_lib, error.code may be a number or a string
            try:
                err_num = int(error.code)
            except ValueError:
                err_num = -1
            if err_num == 60:                                                   # API did not finish on time
                # even if the ZAPI reports a timeout error, it does it after the command completed
                msg = MSGS['dl_completed_slowly']
            elif err_num == 502 and not self.parameters['fail_on_502_error']:   # Bad Gateway
                # ONTAP proxy breaks the connection after 5 minutes, we can assume the download is progressing slowly
                msg = MSGS['dl_in_progress']
            else:
                self.module.fail_json(msg='Error running command %s: %s' % (command, to_native(error)),
                                      exception=traceback.format_exc())
        except netapp_utils.zapi.etree.XMLSyntaxError as error:
            self.module.fail_json(msg='Error decoding output from command %s: %s' % (command, to_native(error)),
                                  exception=traceback.format_exc())

        if output is not None:
            # command completed, check for success
            status = output.get_attr('status')
            cli_output = output.get_child_content('cli-output')
            if status is None or status != 'passed' or cli_output is None or cli_output == "":
                if status is None:
                    extra_info = "'status' attribute missing"
                elif status != 'passed':
                    extra_info = "check 'status' value"
                else:
                    extra_info = 'check console permissions'
                self.module.fail_json(msg='unable to download package from %s: %s.  Received: %s' %
                                      (self.parameters['package_url'], extra_info, output.to_string()))

            if cli_output is not None:
                if cli_output.startswith('Error:') or \
                        'Failed to download package from' in cli_output:
                    self.module.fail_json(msg='failed to download package from %s: %s' % (self.parameters['package_url'], cli_output))
                msg += "  Extra info: %s" % cli_output

        return msg

    def download_sp_image(self):
        fetch_package = netapp_utils.zapi.NaElement('system-image-fetch-package')
        fetch_package.add_new_child('node', self.parameters['node'])
        fetch_package.add_new_child('package', self.parameters['package_url'])
        if self.parameters.get('rename_package'):
            fetch_package.add_new_child('rename-package', self.parameters['rename_package'])
        if self.parameters.get('replace_package'):
            fetch_package.add_new_child('replace-package', str(self.parameters['replace_package']))
        try:
            self.server.invoke_successfully(fetch_package, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching system image package from %s: %s'
                                      % (self.parameters['package_url'], to_native(error)),
                                  exception=traceback.format_exc())

    def download_sp_image_progress(self):
        progress = netapp_utils.zapi.NaElement('system-image-update-progress-get')
        progress.add_new_child('node', self.parameters['node'])
        progress_info = dict()
        try:
            result = self.server.invoke_successfully(progress, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching system image package download progress: %s'
                                      % (to_native(error)), exception=traceback.format_exc())
        if result.get_child_by_name('phase'):
            progress_info['phase'] = result.get_child_content('phase')
        else:
            progress_info['phase'] = None
        if result.get_child_by_name('exit-message'):
            progress_info['exit_message'] = result.get_child_content('exit-message')
        else:
            progress_info['exit_message'] = None
        if result.get_child_by_name('exit-status'):
            progress_info['exit_status'] = result.get_child_content('exit-status')
        else:
            progress_info['exit_status'] = None
        if result.get_child_by_name('last-message'):
            progress_info['last_message'] = result.get_child_content('last-message')
        else:
            progress_info['last_message'] = None
        if result.get_child_by_name('run-status'):
            progress_info['run_status'] = result.get_child_content('run-status')
        else:
            progress_info['run_status'] = None
        return progress_info

    def reboot_sp(self):
        reboot = netapp_utils.zapi.NaElement('service-processor-reboot')
        reboot.add_new_child('node', self.parameters['node'])
        try:
            self.server.invoke_successfully(reboot, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error rebooting service processor: %s'
                                      % (to_native(error)),
                                  exception=traceback.format_exc())

    def download_sp_firmware(self):
        if self.parameters.get('reboot_sp'):
            self.reboot_sp()
        self.download_sp_image()
        progress = self.download_sp_image_progress()
        # progress only show the current or most recent update/install operation.
        if progress['phase'] == 'Download':
            while progress['run_status'] is not None and progress['run_status'] != 'Exited':
                time.sleep(10)
                progress = self.download_sp_image_progress()
            if progress['exit_status'] != 'Success':
                self.module.fail_json(msg=progress['exit_message'], exception=traceback.format_exc())
            return MSGS['dl_completed']
        return MSGS['no_action']

    def autosupport_log(self):
        """
        Autosupport log for software_update
        :return:
        """
        results = netapp_utils.get_cserver(self.server)
        cserver = netapp_utils.setup_na_ontap_zapi(module=self.module, vserver=results)
        netapp_utils.ems_log_event("na_ontap_firmware_upgrade", cserver)

    def apply(self):
        """
        Apply action to upgrade firmware
        """
        changed = False
        msg = MSGS['no_action']
        self.autosupport_log()
        firmware_update_progress = dict()
        if self.parameters.get('package_url'):
            if not self.module.check_mode:
                if self.parameters.get('firmware_type') == 'service-processor':
                    msg = self.download_sp_firmware()
                else:
                    msg = self.download_firmware()
            changed = True
        if not self.parameters['force_disruptive_update']:
            # disk_qual, disk, shelf, and ACP are automatically updated in background
            # The SP firmware is automatically updated on reboot
            self.module.exit_json(changed=changed, msg=msg)
        if msg == MSGS['dl_in_progress']:
            # can't force an update if the software is still downloading
            self.module.fail_json(msg="Cannot force update: %s" % msg)
        if self.parameters.get('firmware_type') == 'service-processor':
            # service-processor firmware upgrade
            current = self.firmware_image_get(self.parameters['node'])

            if self.parameters.get('state') == 'present' and current:
                if not self.module.check_mode:
                    if self.sp_firmware_image_update():
                        changed = True
                    firmware_update_progress = self.sp_firmware_image_update_progress_get(self.parameters['node'])
                    while firmware_update_progress.get('is-in-progress') == 'true':
                        time.sleep(25)
                        firmware_update_progress = self.sp_firmware_image_update_progress_get(self.parameters['node'])
                else:
                    # we don't know until we try the upgrade
                    changed = True

        elif self.parameters.get('firmware_type') == 'shelf':
            # shelf firmware upgrade
            if self.parameters.get('shelf_module_fw'):
                if self.shelf_firmware_required_get():
                    if not self.module.check_mode:
                        changed = self.shelf_firmware_upgrade()
                    else:
                        changed = True
            else:
                if not self.module.check_mode:
                    changed = self.shelf_firmware_upgrade()
                else:
                    # we don't know until we try the upgrade -- assuming the worst
                    changed = True
        elif self.parameters.get('firmware_type') == 'acp':
            # acp firmware upgrade
            if self.acp_firmware_required_get():
                if not self.module.check_mode:
                    self.acp_firmware_upgrade()
                changed = True
        elif self.parameters.get('firmware_type') == 'disk':
            # Disk firmware upgrade
            if self.parameters.get('disk_fw'):
                if self.disk_firmware_required_get():
                    if not self.module.check_mode:
                        changed = self.disk_firmware_upgrade()
                    else:
                        changed = True
            else:
                if not self.module.check_mode:
                    changed = self.disk_firmware_upgrade()
                else:
                    # we don't know until we try the upgrade -- assuming the worst
                    changed = True

        self.module.exit_json(changed=changed, msg='forced update for %s' % self.parameters.get('firmware_type'))


def main():
    """Execute action"""
    community_obj = NetAppONTAPFirmwareUpgrade()
    community_obj.apply()


if __name__ == '__main__':
    main()
